structure DeflateBits :> DeflateBits =
struct

	(* bool <-> int *)
	
	fun bit_to_int true = 1
	  | bit_to_int false = 0

	fun bit_to_char true = #"1"
	  | bit_to_char false = #"0"

	fun least_bit c = if c mod 2 = 1 then true else false

	(* int <-> bool list *)
	
	fun explode_fixed_lsb_first 0 _ = []
	  | explode_fixed_lsb_first len code = 
	  		least_bit code :: explode_fixed_lsb_first (len - 1) (code div 2)

	fun explode_fixed_msb_first len code = rev (explode_fixed_lsb_first len code)
	
	fun to_int_lsb_first bits =
		let
			fun f (b, result) = (2 * result) + (bit_to_int b)
		in
			foldr f 0 bits
		end

	fun to_int_msb_first bits = to_int_lsb_first (rev bits)
	
	fun to_string bits =
		implode (map bit_to_char bits)

	(* power of 2 *)
	
	fun pow2 0 = 1 
	  | pow2 n =
	  		if (n mod 2) = 1 then 2 * pow2 (n - 1)
			else let val p = pow2 (n div 2) in p * p end

	(* self tests *)
	val tests = [
		least_bit 0 = false,
		least_bit 1 = true,
		least_bit 16 = false,
		least_bit 7 = true,
		bit_to_int true = 1,
		bit_to_int false = 0,
		explode_fixed_msb_first 6 15 = [false,false,true,true,true,true],
		to_int_lsb_first [true,true,false] = 3,
		to_string [false,true,true] = "011",
		to_string [] = ""
		]
end
