structure Gzip :> Gzip =
struct

	exception Invalid_gzip_id
	exception Invalid_gzip_method
	
	fun get_header ioinit =
		let
			fun skip_extra false io = io
			  | skip_extra true io =
			  	let 
					val (len, iolen) = DeflateIO.get_fixed 16 io
				in
					DeflateIO.drop_bytes iolen len
				end

			fun opt_zero_string false io = (NONE, io)
			  | opt_zero_string true io = 
			  		(SOME (DeflateIO.take_zero_string io), DeflateIO.drop_zero_string io)
				
			fun skip_header_crc false io = io
			  | skip_header_crc true io = DeflateIO.drop_bits io 16
				 
			fun tuple12 (v1::v2::v3::v4::v5::v6::v7::v8::v9::v10::v11::v12::[]) =
				(v1,v2,v3,v4,v5,v6,v7,v8,v9,v10,v11,v12)
			  | tuple12 _ = raise Match
			
			val (fixed, iofixed) = DeflateList.state_map
				[DeflateIO.get_fixed 8, DeflateIO.get_fixed 8,
				DeflateIO.get_fixed 8,
				DeflateIO.get_fixed 1, DeflateIO.get_fixed 1,
				DeflateIO.get_fixed 1, DeflateIO.get_fixed 1,
				DeflateIO.get_fixed 1, DeflateIO.get_fixed 3,
				DeflateIO.get_fixed 32,
				DeflateIO.get_fixed 8, DeflateIO.get_fixed 8] ioinit
				
			val (id1, id2, cmethod, 
				ftext, fhcrc, fextra, fname, fcomment, _, 
				mtime, xflag, os) = tuple12 fixed

			val (name, ioname) = 
				opt_zero_string (fname <> 0) (skip_extra (fhcrc <> 0) iofixed)

			val (comment, iocomment) = opt_zero_string (fcomment <> 0) ioname

			val iolast = skip_header_crc (fhcrc <> 0) iocomment
		in
			if id1 = 31 andalso id2 = 139 then
				if cmethod = 8 then
					((mtime, name, comment, os), iolast)
				else
					raise Invalid_gzip_method
			else
				raise Invalid_gzip_id
		end

	fun get_trailer ioinit =
		let
			(* warning: overflows if Int.size = 31 ... *)
			val (crc, ionext) = DeflateIO.get_fixed 32 ioinit
			val (isize, iolast) = DeflateIO.get_fixed 32 ionext
		in
			((crc,isize),iolast)
		end
		
	fun unzip_io io =
		let
			val (header, ioheader) = get_header io
			val (ioresult) = Deflate.inflate ioheader
			(*val (trailer, iolast) = get_trailer ioresult*)
		in
			ioresult
		end
		 
	fun unzip_string input =
		implode (map chr (DeflateIO.output 
			(unzip_io (DeflateIO.from_input (map ord (explode input))))))
	
end
